@@ -1,4 +1,5 @@
-# Time-stamp: <2023-08-02 17:32:47 Tao Liu>
+# Time-stamp: <2024-05-15 19:40:45 Tao Liu>
+
 
 """Description: Main HMMR command
 
@@ -16,6 +17,8 @@ import sys
 import gc
 import numpy as np
 import json
+import csv
+import tempfile
 from hmmlearn import hmm
 #from typing import Sized
 
@@ -25,13 +28,13 @@ from hmmlearn import hmm
 from MACS3.Utilities.Constants import *
 from MACS3.Utilities.OptValidator import opt_validate_hmmratac
 from MACS3.IO.PeakIO import PeakIO
-from MACS3.IO.PeakIO import BroadPeakIO
 from MACS3.IO.Parser import BAMPEParser, BEDPEParser #BAMaccessor
 from MACS3.Signal.HMMR_EM import HMMR_EM
 from MACS3.Signal.HMMR_Signal_Processing import generate_weight_mapping, generate_digested_signals, extract_signals_from_regions
 from MACS3.Signal.HMMR_HMM import hmm_training, hmm_predict, hmm_model_init, hmm_model_save
 from MACS3.Signal.Region import Regions
 from MACS3.Signal.BedGraph import bedGraphTrackI
+from MACS3.IO.BedGraphIO import bedGraphIO
 
 #from MACS3.IO.BED import BEDreader # this hasn't been implemented yet.
 
@@ -59,15 +62,21 @@ def run( args ):
     mono_bdgfile = os.path.join( options.outdir, options.name+"_digested_mono.bdg" )
     di_bdgfile = os.path.join( options.outdir, options.name+"_digested_di.bdg" )
     tri_bdgfile = os.path.join( options.outdir, options.name+"_digested_tri.bdg" )
+    
     training_region_bedfile = os.path.join( options.outdir, options.name+"_training_regions.bed" )
     training_datafile = os.path.join( options.outdir, options.name+"_training_data.txt" )
     training_datalengthfile = os.path.join( options.outdir, options.name+"_training_lengths.txt" )
+
     hmm_modelfile = os.path.join( options.outdir, options.name+"_model.json" )
+
     open_state_bdgfile = os.path.join( options.outdir, options.name+"_open.bdg" )
     nuc_state_bdgfile = os.path.join( options.outdir, options.name+"_nuc.bdg" )
     bg_state_bdgfile = os.path.join( options.outdir, options.name+"_bg.bdg" )
+
     states_file = os.path.join( options.outdir, options.name+"_states.bed" )
-    accessible_file = os.path.join( options.outdir, options.name+"_accessible_regions.gappedPeak" )
+
+    accessible_file = os.path.join( options.outdir, options.name+"_accessible_regions.narrowPeak" )
+
     cutoffanalysis_file = os.path.join( options.outdir, options.name+"_cutoff_analysis.tsv" )
     
     #############################################
@@ -156,18 +165,17 @@ def run( args ):
 
     # save three types of signals if needed
     if options.save_digested:
-        fhd = open(short_bdgfile,"w")
-        digested_atac_signals[ 0 ].write_bedGraph(fhd, "short","short")
-        fhd.close()
-        fhd = open(mono_bdgfile,"w")
-        digested_atac_signals[ 1 ].write_bedGraph(fhd, "mono","mono")
-        fhd.close()
-        fhd = open(di_bdgfile,"w")
-        digested_atac_signals[ 2 ].write_bedGraph(fhd, "di","di")
-        fhd.close()
-        fhd = open(tri_bdgfile,"w")
-        digested_atac_signals[ 3 ].write_bedGraph(fhd, "tri","tri")
-        fhd.close()        
+        bdgshort = BedGraphIO( short_bdgfile, data = digested_atac_signals[ 0 ] )
+        bdgshort.write_bedGraph("short","short")
+
+        bdgmono = BedGraphIO( mono_bdgfile, data = digested_atac_signals[ 1 ] )
+        bdgmono.write_bedGraph("mono", "mono")
+
+        bdgdi = BedGraphIO( di_bdgfile, data = digested_atac_signals[ 2 ] )
+        bdgdi.write_bedGraph("di", "di")
+
+        bdgtri = BedGraphIO( tri_bdgfile, data = digested_atac_signals[ 3 ] )
+        bdgtri.write_bedGraph("tri", "tri")
 
     minlen = int(petrack.average_template_length)
     # if options.pileup_short is on, we pile up only the short fragments to identify training 
@@ -175,18 +183,13 @@ def run( args ):
     if options.pileup_short:
         options.info( f"#  Pile up ONLY short fragments" )
         fc_bdg = digested_atac_signals[ 0 ]
-        (sum_v, n_v, max_v, min_v, mean_v, std_v) = fc_bdg.summary()
-        print(sum_v, n_v, max_v, min_v, mean_v, std_v)
-        options.info( f"#  Convert pileup to fold-change over average signal" )
-        fc_bdg.apply_func(lambda x: x/mean_v)
     else:
         options.info( f"#  Pile up all fragments" )
         fc_bdg = petrack.pileup_bdg( [1.0,], baseline_value = 0 )
-        (sum_v, n_v, max_v, min_v, mean_v, std_v) = fc_bdg.summary()
-        options.info( f"#  Convert pileup to fold-change over average signal" )
-        fc_bdg.apply_func(lambda x: x/mean_v)
-       
-        
+    (sum_v, n_v, max_v, min_v, mean_v, std_v) = fc_bdg.summary()
+    options.info( f"#  Convert pileup to fold-change over average signal" )
+    fc_bdg.apply_func(lambda x: x/mean_v)
+
     # if cutoff_analysis only, generate and save the report and quit
     if options.cutoff_analysis_only:
         # we will run cutoff analysis only and quit
@@ -195,11 +198,10 @@ def run( args ):
 
         # Let MACS3 do the cutoff analysis to help decide the lower and upper cutoffs
         with open(cutoffanalysis_file, "w") as ofhd_cutoff:
-            ofhd_cutoff.write( fc_bdg.cutoff_analysis( min_length=minlen, max_gap=options.hmm_training_flanking, max_score = 1000 ) )
+            ofhd_cutoff.write( fc_bdg.cutoff_analysis( min_length=minlen, max_gap=options.hmm_training_flanking, max_score = options.cutoff_analysis_max, steps=options.cutoff_analysis_steps ) )
         #raise Exception("Cutoff analysis only.")
         sys.exit(1)
-        
-        
+
     #############################################
     # 3. Define training set by peak calling
     #############################################
@@ -242,7 +244,7 @@ def run( args ):
         
         # Let MACS3 do the cutoff analysis to help decide the lower and upper cutoffs
         with open(cutoffanalysis_file, "w") as ofhd_cutoff:
-            ofhd_cutoff.write( fc_bdg.cutoff_analysis( min_length=minlen, max_gap=options.hmm_training_flanking, max_score = 1000 ) )
+            ofhd_cutoff.write( fc_bdg.cutoff_analysis( min_length=minlen, max_gap=options.hmm_training_flanking, max_score = options.cutoff_analysis_max ) )
             
         # we will check if anything left after filtering
         if peaks.total > options.hmm_maxTrain:
@@ -274,15 +276,16 @@ def run( args ):
     # 4. Train HMM
     #############################################
     # if model file is provided, we skip this step
+    # include options.hmm_type and make it backwards compatible, if no hmm_type default is gaussian
     if options.hmm_file:
         options.info( f"#4 Load Hidden Markov Model from given model file")
-        hmm_model, i_open_region, i_background_region, i_nucleosomal_region, options.hmm_binsize = hmm_model_init( options.hmm_file )
+        hmm_model, i_open_region, i_background_region, i_nucleosomal_region, options.hmm_binsize, options.hmm_type = hmm_model_init( options.hmm_file )
     else:
         options.info( f"#4 Train Hidden Markov Model with Multivariate Gaussian Emission" )
 
         # extract signals within peak using the given binsize
         options.info( f"#  Extract signals in training regions with bin size of {options.hmm_binsize}")
-        [ training_bins, training_data, training_data_lengths ] = extract_signals_from_regions( digested_atac_signals, training_regions, binsize = options.hmm_binsize )
+        [ training_bins, training_data, training_data_lengths ] = extract_signals_from_regions( digested_atac_signals, training_regions, binsize = options.hmm_binsize, hmm_type = options.hmm_type )
 
         if options.save_train:
             f = open( training_datafile, "w" )
@@ -299,12 +302,15 @@ def run( args ):
 
         options.info( f"#  Use Baum-Welch algorithm to train the HMM")
 
-        hmm_model = hmm_training( training_data, training_data_lengths, random_seed = options.hmm_randomSeed, covar="full" )
+        hmm_model = hmm_training( training_data, training_data_lengths, random_seed = options.hmm_randomSeed, hmm_type = options.hmm_type )
 
         options.info( f"#   HMM converged: {hmm_model.monitor_.converged}")
 
         # label hidden states
-        means_sum = np.sum( hmm_model.means_, axis=1 )
+        if options.hmm_type == "gaussian":
+            means_sum = np.sum( hmm_model.means_, axis=1 )
+        if options.hmm_type == "poisson":
+            means_sum = np.sum( hmm_model.lambdas_, axis=1 )
 
         # first, the state with the highest overall emission is the open state
         i_open_region = np.where( means_sum == max(means_sum) )[0][0]
@@ -317,7 +323,7 @@ def run( args ):
 
         # write hmm into model file
         options.info( f"#  Write HMM parameters into JSON: {hmm_modelfile}")
-        hmm_model_save( hmm_modelfile, hmm_model, options.hmm_binsize, i_open_region, i_nucleosomal_region, i_background_region )
+        hmm_model_save( hmm_modelfile, hmm_model, options.hmm_binsize, i_open_region, i_nucleosomal_region, i_background_region, options.hmm_type )
         
         # if --modelonly option provided, exit script after hmm model is saved 
         if options.hmm_modelonly:
@@ -342,11 +348,19 @@ def run( args ):
     options.info(  "#       {0:>10s}-> {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g}".format(assignments[0], hmm_model.transmat_[0]) )
     options.info(  "#       {0:>10s}-> {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g}".format(assignments[1], hmm_model.transmat_[1]) )
     options.info(  "#       {0:>10s}-> {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g}".format(assignments[2], hmm_model.transmat_[2]) )
-    options.info( f"#   HMM Emissions (mean): ")
-    options.info(  "#                    {0[0]:>10s} {0[1]:>10s} {0[2]:>10s} {0[3]:>10s}".format( ["short", "mono", "di", "tri"] ) )
-    options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[0], hmm_model.means_[0]) )
-    options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[1], hmm_model.means_[1]) )
-    options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[2], hmm_model.means_[2]) )
+    
+    if options.hmm_type == 'gaussian':
+        options.info( f"#   HMM Emissions (means): ")
+        options.info(  "#                    {0[0]:>10s} {0[1]:>10s} {0[2]:>10s} {0[3]:>10s}".format( ["short", "mono", "di", "tri"] ) )
+        options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[0], hmm_model.means_[0]) )
+        options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[1], hmm_model.means_[1]) )
+        options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[2], hmm_model.means_[2]) )
+    if options.hmm_type == 'poisson':
+        options.info( f"#   HMM Emissions (lambdas): ")
+        options.info(  "#                    {0[0]:>10s} {0[1]:>10s} {0[2]:>10s} {0[3]:>10s}".format( ["short", "mono", "di", "tri"] ) )
+        options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[0], hmm_model.lambdas_[0]) )
+        options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[1], hmm_model.lambdas_[1]) )
+        options.info(  "#       {0:>10s}:  {1[0]:>10.4g} {1[1]:>10.4g} {1[2]:>10.4g} {1[3]:>10.4g}".format(assignments[2], hmm_model.lambdas_[2]) )
     
 
 #############################################
@@ -365,11 +379,12 @@ def run( args ):
     options.info( f"#  We expand the candidate regions with {options.hmm_training_flanking} and merge overlap" )
     candidate_regions.expand( options.hmm_training_flanking )
     candidate_regions.merge_overlap()
+    options.info( f"#   after expanding and merging, we have {candidate_regions.total} candidate regions" )
     
     # remove peaks overlapping with blacklisted regions
     if options.blacklist:
         candidate_regions.exclude( blacklist_regions )
-        options.info( f"#  after removing those overlapping with provided blacklisted regions, we have {candidate_regions.total} left" )
+        options.info( f"#   after removing those overlapping with provided blacklisted regions, we have {candidate_regions.total} left" )
 
     # extract signals
     options.info( f"#  Extract signals in candidate regions and decode with HMM")
@@ -380,21 +395,35 @@ def run( args ):
     # Note: we implement in a way that we will decode the candidate regions 10000 regions at a time so 1. we can make it running in parallel in the future; 2. we can reduce the memory usage.
     options.info( f"#  Use HMM to predict states")
     n = 0
-    predicted_proba = []
-    candidate_bins = []
+
+    # we create a temporary file to save the proba predicted from hmm
+    predicted_proba_file = tempfile.TemporaryFile(mode="w+b")
+
     while candidate_regions.total != 0:
         n += 1
-        cr = candidate_regions.pop( options.decoding_steps )
-        options.info( "#    decoding %d..." % ( n*options.decoding_steps ) )        
-        [ cr_bins, cr_data, cr_data_lengths ] = extract_signals_from_regions( digested_atac_signals, cr, binsize = options.hmm_binsize )
-        #options.info( "#     extracted signals in the candidate regions")
-        candidate_bins.extend( cr_bins )
-        #options.info( "#     saving information regarding the candidate regions")        
-        predicted_proba.extend( hmm_predict( cr_data, cr_data_lengths, hmm_model ) )
-        #options.info( "#     prediction done")
+        # we get DECODING_STEPS number of candidate regions first
+        cr = candidate_regions.pop(options.decoding_steps)
+        options.info("#    decoding %d..." % (n * options.decoding_steps))
+
+        # then extrac data from digested signals, create cr_bins, cr_data, and cr_data_lengths
+        [cr_bins, cr_data, cr_data_lengths] = extract_signals_from_regions( digested_atac_signals, cr, binsize = options.hmm_binsize, hmm_type = options.hmm_type )
+        #options.debug( "#     extract_signals_from_regions complete")
+
+        prob_data = hmm_predict(cr_data, cr_data_lengths, hmm_model)
+        assert len(prob_data) == len(cr_bins)
+        for i in range(len(prob_data)):
+            predicted_proba_file.write(b"%s,%d" % cr_bins[i])
+            predicted_proba_file.write(b",%f,%f,%f\n" % tuple(prob_data[i]))
+
+        cr_data = []
+        cr_data_lengths = []
+        cr_bins = []
+        prob_data = []
         gc.collect()
 
-        
+    predicted_proba_file.seek(0) # reset
+    options.info( f"# predicted_proba files written...")
+
 #############################################
 # 6. Output - add to OutputWriter
 #############################################
@@ -412,99 +441,132 @@ def run( args ):
         open_state_bdg_fhd = open( open_state_bdgfile, "w" )
         nuc_state_bdg_fhd = open( nuc_state_bdgfile, "w" )
         bg_state_bdg_fhd = open( bg_state_bdgfile, "w" )
-        save_proba_to_bedGraph( candidate_bins, predicted_proba, options.hmm_binsize, open_state_bdg_fhd, nuc_state_bdg_fhd, bg_state_bdg_fhd, i_open_region, i_nucleosomal_region, i_background_region )
+        save_proba_to_bedGraph( predicted_proba_file, options.hmm_binsize, open_state_bdg_fhd, nuc_state_bdg_fhd, bg_state_bdg_fhd, i_open_region, i_nucleosomal_region, i_background_region )
+        predicted_proba_file.seek(0) # reset
         open_state_bdg_fhd.close()
         nuc_state_bdg_fhd.close()
         bg_state_bdg_fhd.close()
+        options.info( f"# finished writing proba_to_bedgraph")
     
-    # Generate states path:
-    states_path = generate_states_path( candidate_bins, predicted_proba, options.hmm_binsize, i_open_region, i_nucleosomal_region, i_background_region )
-    
+    # # Generate states path:
+    states_path = generate_states_path( predicted_proba_file, options.hmm_binsize, i_open_region, i_nucleosomal_region, i_background_region )
+    options.info( f"# finished generating states path")
+    predicted_proba_file.close()          #kill the temp file
     # Save states path if needed
     # PS: we need to implement extra feature to include those regions NOT in candidate_bins and assign them as 'background state'.
     if options.save_states:
         options.info( f"# Write states assignments in a BED file: {options.name}_states.bed" )
-        f = open( states_file, "w" )
-        save_states_bed( states_path, f )
-        f.close()
+        with open( states_file, "w" ) as f:
+            save_states_bed( states_path, f )
 
-    options.info( f"# Write accessible regions in a gappedPeak file: {options.name}_accessible_regions.gappedPeak")
-    ofhd = open( accessible_file, "w" )
-    save_accessible_regions( states_path, ofhd, options.openregion_minlen )
-    ofhd.close()
+    options.info( f"# Write accessible regions in a narrowPeak file: {options.name}_accessible_regions.narrowPeak")
+    with open( accessible_file, "w" ) as ofhd:
+        save_accessible_regions( states_path, ofhd, options.openregion_minlen, fc_bdg )
 
-def save_proba_to_bedGraph( candidate_bins, predicted_proba, binsize, open_state_bdg_file, nuc_state_bdg_file, bg_state_bdg_file, i_open, i_nuc, i_bg ):
-    open_state_bdg = bedGraphTrackI( baseline_value = 0 )
-    nuc_state_bdg = bedGraphTrackI( baseline_value = 0 )
-    bg_state_bdg = bedGraphTrackI( baseline_value = 0 )
+    options.info( f"# Finished")
 
+def save_proba_to_bedGraph( predicted_proba_file, binsize, open_state_bdg_file, nuc_state_bdg_file, bg_state_bdg_file, i_open, i_nuc, i_bg ):
+
+    open_state_bdg_file = bedGraphIO( open_state_bdg_file )
+    nuc_state_bdg_file = bedGraphIO( nuc_state_bdg_file )
+    bg_state_bdg_file = bedGraphIO( bg_state_bdg_file )
+
+    open_state_bdg = open_state_bdg_file.data
+    nuc_state_bdg = nuc_state_bdg_file.data
+    bg_state_bdg = bg_state_bdg_file.data
+    
     prev_chrom_name = None
     prev_bin_end = None
-    for l in range(len(predicted_proba)):
-        # note that any region not in the candidate bins will be
-        # treated as absolutely the background
-        chrname = candidate_bins[l][0]
-        end_pos = candidate_bins[l][1]
+
+    for pp_line in predicted_proba_file:
+        pp_data = pp_line.strip().split(b',')
+
+        chrname = pp_data[0]
+        end_pos = int(pp_data[1])
         start_pos = end_pos - binsize
+        pp_open = float(pp_data[i_open+2])
+        pp_nuc = float(pp_data[i_nuc+2])
+        pp_bg = float(pp_data[i_bg+2])
+
         if chrname != prev_chrom_name:
-            prev_chrom_name = chrname
-            # add the first region as background
+            # we start a new chromosome
             if start_pos > 0:
+                # add the first unannotated region as background
                 open_state_bdg.add_loc( chrname, 0, start_pos, 0.0 )
                 nuc_state_bdg.add_loc( chrname, 0, start_pos, 0.0 )
                 bg_state_bdg.add_loc( chrname, 0, start_pos, 1.0 )
-                prev_bin_end = start_pos
-            elif start_pos == 0:
-                # if start_pos == 0, then the first bin has to be assigned, we set prev_bin_end as 0 
-                prev_bin_end = 0
-        # now check if the prev_bin_end is start_pos, if not, add a gap of background
-        if prev_bin_end < start_pos:
-            open_state_bdg.add_loc( chrname, prev_bin_end, start_pos, 0.0 )
-            nuc_state_bdg.add_loc( chrname, prev_bin_end, start_pos, 0.0 )
-            bg_state_bdg.add_loc( chrname, prev_bin_end, start_pos, 1.0 )
+            prev_chrom_name = chrname
+        else:
+            # now check if the prev_bin_end is start_pos, if not, add a gap of background
+            if prev_bin_end < start_pos:
+                open_state_bdg.add_loc( chrname, prev_bin_end, start_pos, 0.0 )
+                nuc_state_bdg.add_loc( chrname, prev_bin_end, start_pos, 0.0 )
+                bg_state_bdg.add_loc( chrname, prev_bin_end, start_pos, 1.0 )
 
-        open_state_bdg.add_loc( chrname, start_pos, end_pos, predicted_proba[l][ i_open ] )
-        nuc_state_bdg.add_loc( chrname, start_pos, end_pos, predicted_proba[l][ i_nuc ] )
-        bg_state_bdg.add_loc( chrname, start_pos, end_pos, predicted_proba[l][ i_bg ] )
-        prev_bin_end = start_pos
-
-    open_state_bdg.write_bedGraph( open_state_bdg_file, "Open States", "Likelihoods of being Open States" )
-    nuc_state_bdg.write_bedGraph( nuc_state_bdg_file, "Nucleosomal States", "Likelihoods of being Nucleosomal States" )
-    bg_state_bdg.write_bedGraph( bg_state_bdg_file, "Background States", "Likelihoods of being Background States" )
+        open_state_bdg.add_loc( chrname, start_pos, end_pos, pp_open )
+        nuc_state_bdg.add_loc( chrname, start_pos, end_pos, pp_nuc )
+        bg_state_bdg.add_loc( chrname, start_pos, end_pos, pp_bg )
+        prev_bin_end = end_pos
+        
+    open_state_bdg_file.write_bedGraph( "Open States", "Likelihoods of being Open States", trackline = False )
+    nuc_state_bdg_file.write_bedGraph( "Nucleosomal States", "Likelihoods of being Nucleosomal States", trackline = False )
+    bg_state_bdg_file.write_bedGraph( "Background States", "Likelihoods of being Background States", trackline = False )
     return
 
 def save_states_bed( states_path, states_bedfile ):
     # we do not need to output background state. 
     for l in range( len( states_path ) ):
         if states_path[l][3] != "bg":
-            states_bedfile.write( "%s\t%s\t%s\t%s\n" % states_path[l] )
+            states_bedfile.write( "%s\t" % states_path[l][0].decode() )
+            states_bedfile.write( "%d\t%d\t%s\n" % states_path[l][1:] )
     return
 
-def generate_states_path(candidate_bins, predicted_proba, binsize, i_open_region, i_nucleosomal_region, i_background_region):
+def generate_states_path(predicted_proba_file, binsize, i_open, i_nuc, i_bg):
+    # predicted_proba_file is a temporary file
     ret_states_path = []
     labels_list = ["open", "nuc", "bg"]
-    start_pos = candidate_bins[0][1] - binsize
-    for l in range(1, len(predicted_proba)):
-        chromosome = candidate_bins[l][0].decode()
-        prev_open, prev_nuc, prev_bg = predicted_proba[l-1][i_open_region], predicted_proba[l-1][i_nucleosomal_region], predicted_proba[l-1][i_background_region]
-        curr_open, curr_nuc, curr_bg = predicted_proba[l][i_open_region], predicted_proba[l][i_nucleosomal_region], predicted_proba[l][i_background_region]
-        
-        label_prev = labels_list[max((prev_open, 0), (prev_nuc, 1), (prev_bg, 2), key=lambda x: x[0])[1]]
-        label_curr = labels_list[max((curr_open, 0), (curr_nuc, 1), (curr_bg, 2), key=lambda x: x[0])[1]]
 
-        if candidate_bins[l-1][0] == candidate_bins[l][0]:  # if we are looking at the same chromosome ...
-            if label_prev != label_curr:
-                end_pos = candidate_bins[l-1][1]
-                ret_states_path.append((chromosome, start_pos, end_pos, label_prev))
-                start_pos = candidate_bins[l][1] - binsize
-            elif l == len(predicted_proba) - 1:
-                end_pos = candidate_bins[l][1]
-                ret_states_path.append((chromosome, start_pos, end_pos, label_prev))
+    prev_chrom_name = None
+    prev_bin_end = None
+    prev_label = None
+    
+    for pp_line in predicted_proba_file:
+        pp_data = pp_line.strip().split(b',')
+
+        chrname = pp_data[0]
+        end_pos = int(pp_data[1])
+        start_pos = end_pos - binsize
+        pp_open = float(pp_data[i_open+2])
+        pp_nuc = float(pp_data[i_nuc+2])
+        pp_bg = float(pp_data[i_bg+2])
+
+        # find the best state as label
+        label = labels_list[max((pp_open, 0), (pp_nuc, 1), (pp_bg, 2), key=lambda x: x[0])[1]]
+        
+        if chrname != prev_chrom_name:
+            # we start a new chromosome
+            if start_pos > 0:
+                # add the first unannotated region as background
+                ret_states_path.append((chrname, 0, start_pos, "bg"))
+            ret_states_path.append((chrname, start_pos, end_pos, label))
+            prev_label = label
+            prev_chrom_name = chrname
         else:
-            start_pos = candidate_bins[l][1] - binsize
+            # now check if the prev_bin_end is start_pos, if not, add a gap of background
+            if prev_bin_end < start_pos:
+                ret_states_path.append((chrname, prev_bin_end, start_pos, "bg"))
+                prev_label = "bg"
+            # same label, just extend
+            if label == prev_label:
+                ret_states_path[-1] = (ret_states_path[-1][0], ret_states_path[-1][1], end_pos, label)
+            else:
+                ret_states_path.append((chrname, start_pos, end_pos, label))
+                prev_label = label
+
+        prev_bin_end = end_pos
     return ret_states_path
 
-def save_accessible_regions(states_path, accessible_region_file, openregion_minlen):
+def save_accessible_regions(states_path, accessible_region_file, openregion_minlen, bdgscore):
     # Function to add regions to the list
     def add_regions(i, regions):
         for j in range(i, i+3):
@@ -518,31 +580,18 @@ def save_accessible_regions(states_path, accessible_region_file, openregion_minl
     for i in range(len(states_path)-2):
         if (states_path[i][3] == 'nuc' and states_path[i+1][3] == 'open' and states_path[i+2][3] == 'nuc' and 
             states_path[i][2] == states_path[i+1][1] and states_path[i+1][2] == states_path[i+2][1] and 
-            states_path[i+1][2] - states_path[i+1][1] > openregion_minlen):
+            states_path[i+2][2] - states_path[i][1] > openregion_minlen): # require nuc-open-nuc entire region start/endpos > openregion_minlen
             accessible_regions = add_regions(i, accessible_regions)
-
-    # Group states by region
-    list_of_groups = []
-    one_group = [accessible_regions[0]]
-    for j in range(1, len(accessible_regions)):
-        if accessible_regions[j][1] == accessible_regions[j-1][2]:
-            one_group.append(accessible_regions[j])
-        else:
-            list_of_groups.append(one_group)
-            one_group = [accessible_regions[j]]
-    accessible_regions = list_of_groups
+    
+    # remove 'nuc' regions: 
+    accessible_regions = [tup for tup in accessible_regions if tup[3] != 'nuc']
 
     # Generate broadpeak object
-    broadpeak = BroadPeakIO()
+    openpeak = PeakIO()
     for region in accessible_regions[:-1]:
-        block_num = sum('open' in tup for tup in region)
-        block_sizes = ','.join(str(region[j][2] - region[j][1]) for j in range(1, len(region) - 1, 2))
-        block_starts = ','.join(str(region[j][1] - region[0][1]) for j in range(1, len(region) - 1, 2))
-        broadpeak.add(bytes(region[1][0], encoding="raw_unicode_escape"), region[0][1], region[-1][2],
-                      thickStart=bytes(str(region[1][1]), encoding="raw_unicode_escape"),
-                      thickEnd=bytes(str(region[-2][2]), encoding="raw_unicode_escape"),
-                      blockNum=block_num,
-                      blockSizes=bytes(block_sizes, encoding="raw_unicode_escape"),
-                      blockStarts=bytes(block_starts, encoding="raw_unicode_escape"))
-    broadpeak.write_to_gappedPeak(accessible_region_file)
+        openpeak.add(chromosome=region[0], start=region[1], end=region[2]) 
+
+    # refine peak summit and score using bedGraphTrackI with scores
+    openpeak = bdgscore.refine_peaks( openpeak )
+    openpeak.write_to_narrowPeak(accessible_region_file)
     return
